from EnteroTyper.bin.accessories import run_subprocess, dependency_check, check_all_dependencies


def test_run_subprocess():
    assert run_subprocess('echo test', get_stdout=True) == 'test'
    assert run_subprocess('echo test', get_stdout=False) is None


def test_dependency_check():
    assert dependency_check('echo') is True
    assert dependency_check('ls') is True
    assert dependency_check('DefinitelyNotARealProgram__123') is False


def test_check_all_dependencies():
    assert check_all_dependencies(['echo', 'ls']) is True
    assert check_all_dependencies(['Definitely__Fake1', 'Definitely__Fake2']) is False



import os
import logging
import pandas as pd
import multiprocessing
from copy import deepcopy
from tqdm import tqdm
from pathlib import Path
from EnteroTyper.bin.compare_sequences_types import call_sequence_comparison
from EnteroTyper.bin.typer import get_database_files, generate_cgmlst_report
from EnteroTyper.bin.accessories import run_subprocess


def sequence_concatenation_pipeline(targets: list, database: Path, outdir: Path):
    targets = [Path(target) for target in targets]
    database_files = get_database_files(database=database)
    os.makedirs(str(outdir), exist_ok=False)
    logging.debug(f"Created directory {outdir}")

    report_dict = {}
    for i, target in enumerate(targets):
        target_name = target.with_suffix("").name + f'_{i}'
        report_dict[target_name] = target

    logging.info("Generating FASTA files from reports")
    df_params = [(database_file, outdir, report_dict) for database_file in database_files]
    with multiprocessing.Pool(multiprocessing.cpu_count() - 1) as p:
        p.starmap(write_fasta, df_params)

    logging.info("Aligning FASTA files with MUSCLE")
    fasta_files = list(outdir.glob("*.fasta"))
    with multiprocessing.Pool(multiprocessing.cpu_count() - 1) as p:
        p.map(call_muscle, fasta_files)

    # Grab newly aligned fasta files
    aligned_fasta_files = list(outdir.glob("*.align.fasta"))

    # Create multifasta with fasconcat
    logging.info("Concatenating sequences")
    n_processes = int(multiprocessing.cpu_count() / 8)
    sample_ids = list(report_dict.keys())
    outfile = concatenate_sequence_directory(sample_ids=sample_ids, sequence_directory=outdir, n_processes=n_processes,
                                             outdir=outdir)
    logging.info(f"Concatenated sequences available at {outfile}")

    # Remove remaining fasta files
    [os.remove(str(fasta)) for fasta in aligned_fasta_files]

    # Sequence type comparison
    logging.info("Conducting sequence type comparisons")
    sequence_type_report_list = []
    for target in tqdm(targets):
        df = pd.read_csv(target, sep='\t')
        cgmlst_allele_report = generate_cgmlst_report(df=df, outdir=outdir,
                                                      sample_name=target.name.rsplit("_", 1)[0])
        sequence_type_report_list.append(cgmlst_allele_report)

    call_sequence_comparison(targets=sequence_type_report_list, outdir=outdir / 'sequence_type_comparisons')

    # Cleanup reports
    for report in sequence_type_report_list:
        try:
            os.remove(str(report))
        except FileNotFoundError:
            continue


def get_top_qseq(df: pd.DataFrame, locus: str) -> str:
    locus_df = df[df['locus'] == locus]
    qseq = locus_df['qseq_strand_aware'].iloc[0]
    if qseq == "NA":
        return ""
    return qseq


def read_blast_report(report: Path):
    df = pd.read_csv(report, delimiter="\t")
    return df


def write_fasta(database_file: Path, outdir: Path, report_dict: dict):
    locus = database_file.with_suffix("").name
    locus_length = 0
    logging.debug(f"{locus}...")
    outfile = str(outdir / database_file.with_suffix(".fasta").name)
    with open(outfile, "w") as out:
        for target, report_path in report_dict.items():
            out.write(f">{target}\n")
            df = read_blast_report(report_path)
            qseq = get_top_qseq(df=df, locus=locus)
            if type(qseq) == str:
                locus_length = len(qseq)
            else:
                qseq = "N" * locus_length
            out.write(f"{qseq}\n")


def call_muscle(infile: Path):
    """
    Produces an aligned version of an input FASTA file, then removes the original.
    """
    outfile = infile.with_suffix(".align.fasta")
    cmd = f"muscle -in {infile} -out {outfile} -maxiters 1"
    run_subprocess(cmd, get_stdout=True)
    infile.unlink()


def concatenate_sequence_directory(sample_ids: [str], sequence_directory: Path, n_processes: int, outdir: Path) -> Path:
    """
    Given a sequence directory containing aligned multi-FASTA files, will attempt to concatenate all of the sequences
    into a single file. Minimal example showing input and output (concatenated_sequences.fasta):

    files in directory:
        gene_1.fasta
        gene_2.fasta

    contents of gene1.fasta:
        >sample_1
        ATCG
        >sample2
        ATTT

    contents of gene2.fasta:
        >sample_1
        GGGGAGGGGGTCA

    concatenated_sequences.fasta:
        >sample_1
        ATCGGGGGAGGGGGTCA
        >sample_2
        ATTTNNNNNNNNNNNNN

    - Provided sample_ids list must contain exact matches to headers in FASTA files
    - All sequences in a single FASTA file must be the exact same length; this is done through an aligner like MUSCLE
    - Expectation is that every FASTA file will have the same headers, or a subset of headers present in sample_ids
    """
    concat_seqs_dict = generate_concat_seqs_dict(sample_ids=sample_ids, indir=sequence_directory,
                                                 n_processes=n_processes)
    outfile = write_concat_seqs_dict(concat_seqs_dict=concat_seqs_dict, outdir=outdir)
    return outfile


def write_concat_seqs_dict(concat_seqs_dict: dict, outdir: Path) -> Path:
    outfile = outdir / 'concatenated_sequences.fasta'
    if outfile.exists():
        outfile.unlink()
    outfile_ = open(str(outfile), 'a+')
    for sample_id, sequence in concat_seqs_dict.items():
        outfile_.write(f">{sample_id}\n")
        outfile_.write(f"{sequence}\n")
    outfile_.close()
    return outfile


def generate_concat_seqs_dict(sample_ids: set, indir: Path, n_processes=4) -> dict:
    # Potentially takes a lot of RAM. Stores the concatenated sequences for each sample.
    sequence_storage = {sample_id: "" for sample_id in sample_ids}
    fasta_files = sorted(list(indir.glob("*.fasta")))

    # Set # of concurrent processes to run
    pool = multiprocessing.Pool(processes=n_processes)
    cluster_dicts = [
        pool.apply_async(populate_template_dict, args=(sequence_storage, f, sample_ids)) for f in fasta_files
    ]
    cluster_dicts = [result.get() for result in cluster_dicts]

    # Merge all of the dictionaries into one
    for d in cluster_dicts:
        for cluster, sequence in d.items():
            sequence_storage[cluster] += sequence
    return sequence_storage


def populate_template_dict(template_dict: dict, cluster_file: Path, sample_ids: list):
    with open(str(cluster_file), 'r') as f:
        cluster_dict = deepcopy(template_dict)
        lines = f.readlines()
        cluster_samples = []
        seq_length = 0
        seq_lengths = []
        for line in lines:
            line = line.strip()
            if line.startswith(">"):
                # Add the sequence length to the ongoing tracking list
                if seq_length != 0:
                    seq_lengths.append(seq_length)
                seq_length = 0
                # Remove '>'
                header = line.replace(">", "")
                # Add this header to cluster_samples to keep track of which samples we have sequence for
                cluster_samples.append(header)
            else:
                cluster_dict[header] += line
                seq_length += len(line)
        try:
            assert len(set(seq_lengths)) == 1
        except AssertionError:
            print(f"ERROR: Varying sequence lengths detected in {f.name}")
            print(seq_lengths)
            quit()
        cluster_samples = set(cluster_samples)
        for s in sample_ids:
            if s not in cluster_samples:
                cluster_dict[s] += ('N' * seq_lengths[0])
        return cluster_dict



from EnteroTyper.bin.compare_sequences_types import *


def test_call_sequence_comparison():
    pass


def test_compare_reports():
    pass


def test_read_allele_report():
    pass


def test_compare_keys():
    pass


def test_compare_allele_dicts():
    pass


def test_count_mismatches():
    pass



import shutil
import logging
from subprocess import Popen, PIPE


def run_subprocess(cmd: str, get_stdout: bool = False) -> str:
    if get_stdout:
        p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
        out, err = p.communicate()
        out = out.decode().strip()
        err = err.decode().strip()
        if out != "":
            return out
        elif err != "":
            return err
        else:
            return ""
    else:
        p = Popen(cmd, shell=True)
        p.wait()


def dependency_check(dependency: str) -> bool:
    """
    Checks if a given program is present in the user's $PATH
    :param dependency: String of program name
    :return: True if program is in $PATH, False if not
    """
    check = shutil.which(dependency)
    if check is not None:
        return True
    else:
        return False


def check_all_dependencies(dependencies: [str]) -> bool:
    # Dependency check
    logging.info("Conducting dependency check...")
    dependency_dict = dict()
    dependencies_met = True
    for dependency in dependencies:
        dependency_dict[dependency] = dependency_check(dependency)
    if False in dependency_dict.values():
        dependencies_met = False
        logging.error("ERROR: Cannot locate some dependencies in $PATH...")
        for key, value in dependency_dict.items():
            if not value:
                logging.error(f"Dependency missing: {key}")
        return dependencies_met
    else:
        for key, value in dependency_dict.items():
            logging.debug(f"Dependency {key}: {value}")
        logging.info("Dependencies OK")
        return dependencies_met



import os
import gzip
import shutil
import logging
import multiprocessing
import pandas as pd

from tqdm import tqdm
from multiprocessing import Pool
from pathlib import Path
from EnteroTyper.bin.accessories import run_subprocess

script = os.path.basename(__file__)


def type_sample(input_assembly: Path, database: Path, outdir: Path, create_db: bool, keep_blast: bool,
                sample_name: str = None):
    """
    Function wrapping all of the functionality of the enterobase_typer script. Can be called directly.
    """

    # Output directory creation/validation
    os.makedirs(str(outdir), exist_ok=False)
    logging.debug(f"Created directory {outdir}")

    logging.debug(f"input_assembly: {input_assembly}")
    logging.debug(f"database: {database}")
    logging.debug(f"outdir: {outdir}")
    logging.debug(f"create_db: {create_db}")

    # Call makeblastdb on each database file if create_db=True
    if create_db:
        database_files = makeblastdb_database(database=database)
    else:
        database_files = get_database_files(database=database)

    # Query input_assembly against each loci in the cgMLST database
    df = multiprocess_blastn_call(database_files, input_assembly, outdir)

    # Grab sample name from input_assembly
    if sample_name is None:
        sample_name = input_assembly.with_suffix("").name.replace(".pilon", "")

    # Prepare detailed report
    detailed_report = generate_detailed_report(df=df, outdir=outdir, sample_name=sample_name)

    # Prepare cgMLST report
    cgmlst_allele_report = generate_cgmlst_report(df=df, outdir=outdir, sample_name=sample_name)

    logging.info(f"cgMLST Allele Report: {cgmlst_allele_report}")
    logging.info(f"Detailed Report: {detailed_report}")

    if keep_blast:
        # Move BLASTn files
        move_blastn_files(indir=outdir, outdir=(outdir / 'blastn_output'))
    else:
        # Delete all blastn files
        delete_blastn_files(indir=outdir)
    return detailed_report


def move_blastn_files(indir: Path, outdir: Path):
    """
    Moves all of the BLASTn output to a new folder 'blastn_output'
    """
    os.makedirs(str(outdir), exist_ok=True)
    blastn_files = list(indir.glob("*.BLASTn"))
    for f in blastn_files:
        shutil.move(str(f), str(outdir / f.name))


def delete_blastn_files(indir: Path):
    blastn_files = list(indir.glob("*.BLASTn"))
    [os.remove(str(f)) for f in blastn_files]


def generate_cgmlst_report(df: pd.DataFrame, outdir: Path, sample_name: str) -> Path:
    """
    Generates the cgMLST report along with a transposed version within the out_dir folder
    """
    cgmlst_allele_report = outdir / f"{sample_name}_cgMLST_Allele_Report.tsv"
    cgmlst_df = get_sequence_type(df)
    cgmlst_df_transposed = cgmlst_df.transpose()
    cgmlst_df_transposed.to_csv(cgmlst_allele_report, sep="\t", header=False)
    return cgmlst_allele_report


def generate_detailed_report(df: pd.DataFrame, outdir: Path, sample_name: str) -> Path:
    """
    Generates the detailed report within the out_dir folder
    """
    output_detailed_report = outdir / f"{sample_name}_BLASTn_Detailed_Report.tsv"
    df.to_csv(output_detailed_report, sep="\t", index=None)
    return output_detailed_report


def get_database_files(database: Path, loci_suffix: str = "*.gz"):
    """
    Grabs database files within a provided directory and returns a list of everything present.
    Matches against a given suffix (defaults to *.gz, the expected database file extension)
    """
    database_files = list(database.glob(loci_suffix))
    if len(database_files) > 0:
        return database_files
    else:
        raise EmptyDatabase(f"Could not find database files at {database}. "
                            f"Try re-running the script with the --create_db flag.")


def multiprocess_blastn_call(database_files: list, input_assembly: Path, outdir: Path) -> [pd.DataFrame]:
    """
    Calls blastn on every database file against the input_assembly, reads top hit of results into a dataframe per loci,
    then returns a single dataframe of the combined results
    """
    df_params = [(database_file, input_assembly, outdir) for database_file in database_files]

    with Pool(multiprocessing.cpu_count() - 1) as p:
        df_list = p.starmap(closest_allele_df, df_params)
    df = combine_dataframes(df_list)

    # Sort values
    df = df.sort_values(by=['sseqid'])

    # Get reverse complement of minus strands
    df['qseq_strand_aware'] = df.apply(get_reverse_complement_row, axis=1)

    # Drop extraneous columns
    df = df.drop(['qseq', 'sstrand', 'score'], axis=1)

    # Sort again
    df = df.sort_values(by=['locus'])

    return df


def closest_allele_df(database_file: Path, input_assembly: Path, outdir: Path) -> pd.DataFrame:
    """
    Queries the input_assembly against a given database file, returns the top hit as a dataframe
    """
    database_file = database_file.with_suffix(".blastDB")
    blastn_file, locus_name = call_blastn(database_file=database_file, query_fasta=input_assembly, out_dir=outdir)
    df = parse_blastn(blastn_file)
    df = process_blastn_df(df, locus_name)
    return df


def get_sequence_type(df: pd.DataFrame) -> pd.DataFrame:
    loci = df['locus'].unique()
    sequence_type_dict = {}
    for locus in loci:
        sequence_type_df = df[df['locus'] == locus]
        sequence_type = sequence_type_df['sseqid'].iloc[0]
        if str(sequence_type) != 'nan':
            sequence_type = sequence_type.rsplit('_', 1)[-1]
        else:
            sequence_type = "NA"
        sequence_type_dict[locus] = sequence_type
    df = pd.DataFrame.from_dict(sequence_type_dict, orient='index').transpose()
    return df


def process_blastn_df(df: pd.DataFrame, locus_name: str) -> pd.DataFrame:
    """
`   Sorts and filters the DataFrame according to pident and a newly calculated lratio, evaluates top hits, then
    returns a DataFrame with only the top hit along with a 'hit_type' classification.
    :param df: DataFrame containing parsed *.BLASTn results
    :param locus_name: Name of the locus represented in the *.BLASTn file
    :return: df containing only the top hit from the parsed *.BLASTn
    """
    logging.debug(f"Processing locus: {locus_name}...")
    # Filter by length to slen ratio
    df['lratio'] = df['length'] / df['slen']
    df['locus'] = locus_name

    # Filter junk from DataFrame
    df = df.query("lratio >= 0.90 & pident >= 90.0 & lratio <=1.10")

    # Sort values so the best hits are at the top
    df = df.sort_values(["pident", "lratio"], ascending=False)
    df = df.reset_index(drop=True)

    # TODO: Revisit this and rework it. The hit classification is not great.
    hit_type = None
    if len(df) == 1:
        # CASE 1: PERFECT SINGLE HIT
        if df.pident[0] == 100.0 and df.lratio[0] == 1.0:
            hit_type = "PERFECT_HIT"
    elif len(df) > 1 and df.pident[0] == 100.0 and df.lratio[0] == 1.0:
        # CASE 2: PERFECT DUPLICATE - top two hits both have 100% pident and lratio
        if df.pident[1] == 100.0 and df.lratio[1] == 1.0:
            hit_type = "PERFECT_DUPLICATE"
        # CASE 3: PERFECT HIT w/very close second hit
        elif (0.99 <= df.lratio[1] <= 1.01) or df.pident[1] >= 0.99:
            hit_type = "PERFECT_HIT_WITH_POTENTIAL_PARALOG"
        # CASE 4: PERFECT HIT w/slightly less close second hit
        else:
            hit_type = "PERFECT_HIT"
    elif len(df) == 0:  # no close hits
        hit_type = "NO_MATCH"
        data = []
        data.insert(0, {"qseqid": "NA", "sseqid": "NA", "slen": "NA",
                        "length": "NA", "qstart": "NA", "qend": "NA",
                        "pident": "NA", "score": "NA", "locus": locus_name,
                        "sstrand": "NA", "qseq": "NA", "lratio": "NA"})
        df = pd.concat([pd.DataFrame(data), df], ignore_index=True, sort=False)
    else:
        hit_type = "CLOSEST_HIT"

    df["hit_type"] = hit_type
    return df.head(1)


def call_makeblastdb(db_file: Path):
    """
    Makes a system call to makeblastdb on a given database file. Can handle *.gz, *.fasta, or no suffix.
    :param db_file: Path to database file
    """
    db_name = db_file.with_suffix(".blastDB")
    if db_file.suffix == ".gz":
        cmd = f"gunzip -c {db_file} | makeblastdb -in - -parse_seqids -dbtype nucl -out {db_name} -title {db_name}"
        run_subprocess(cmd, get_stdout=True)
    elif db_file.suffix == ".fasta":
        cmd = f"makeblastdb -in {db_file} -parse_seqids -dbtype nucl -out {db_name} -title {db_name}"
        run_subprocess(cmd, get_stdout=True)
    elif db_file.suffix == "":
        os.rename(str(db_file), str(db_file.with_suffix(".fasta")))
        cmd = f"makeblastdb -in {db_file} -parse_seqids -dbtype nucl -out {db_name} -title {db_name}"
        run_subprocess(cmd, get_stdout=True)
    else:
        logging.debug("Invalid file format provided to call_makeblastdb()")


def makeblastdb_database(database: Path, loci_suffix: str = "*.gz") -> list:
    """
    Calls makeblastdb on every *.gz file in a given directory. Intended to function with the files retrieved via the
    EnterobasePull script (https://github.com/bfssi-forest-dussault/EnterobasePull).
    :param database: Path to database retrieved with EnterobasePull
    :param loci_suffix: Suffix of files to run makeblastdb on
    """
    logging.debug(f"Calling makeblastdb on database at {database}")
    db_files = list(database.glob(loci_suffix))
    for f in tqdm(db_files):
        call_makeblastdb(f)
    return db_files


def get_allele_length(database_file: Path) -> int:
    with gzip.open(database_file, "rt") as handle:
        lines = [line.strip() for line in handle]
        for line in lines:
            if line[0] != ">":
                return len(line)


def call_blastn(database_file: Path, query_fasta: Path, out_dir: Path) -> (Path, str):
    """
    Calls blastn against a query *.fasta file using the provided Enterobase DB target, dumps output into out_dir
    :param database_file: Formatted (makeblastdb) BLAST database file
    :param query_fasta: Sequence to query against database file
    :param out_dir: Path to directory to store output
    :return: BLASTn output (Path) and the base name of the target being searched (str)
    """
    # Dynamically set word_size
    allele_length = get_allele_length(database_file=database_file.with_suffix(".gz"))
    word_size = round(allele_length / 4)

    locus_name = database_file.with_suffix('').name
    reference_name = query_fasta.with_suffix('').name
    out_file = out_dir / Path(reference_name + "." + locus_name + ".BLASTn")
    cmd = f"blastn -query {query_fasta} -db {database_file} -out {out_file} " \
        f"-outfmt '6 qseqid sseqid slen length qstart qend pident score sstrand qseq' -word_size {word_size}"
    run_subprocess(cmd)
    return out_file, locus_name


def combine_dataframes(dfs: [pd.DataFrame]) -> pd.DataFrame:
    """
    Receives a list of DataFrames and concatenates them. They must all have the same header.
    :param dfs: List of DataFrames
    :return: Single concatenated DataFrame
    """
    df = pd.concat(dfs, sort=False)
    return df


def parse_blastn(blastn_file: Path) -> pd.DataFrame:
    """
    Parses *.BLASTn file generated by call_blastn(), then returns the df
    :param blastn_file: file path to *.BLASTn file
    :return: DataFrame contents of *.BLASTn file
    """
    headers = ["qseqid", "sseqid", "slen", "length", "qstart", "qend", "pident", "score", "sstrand", "qseq"]
    df = pd.read_csv(blastn_file, delimiter="\t", names=headers)
    return df


def get_reverse_complement_row(row) -> str:
    """
    Takes DataFrame row and returns the reverse complement of the sequence if the strand is 'minus'
    :param row:
    :return:
    """
    complement_dict = {'A': 'T',
                       'C': 'G',
                       'G': 'C',
                       'T': 'A'}
    sequence = row['qseq'].upper()
    if row['sstrand'] == 'minus':
        reverse_complement = "".join(complement_dict.get(base, base) for base in reversed(sequence))
        return reverse_complement
    else:
        return sequence


class EmptyDatabase(Exception):
    pass



import tempfile
import shutil
from EnteroTyper.bin.bulk_typer import get_sample_name_dict, bulk_sample_typing
from pathlib import Path


def test_get_sample_name_dict():
    # Create mock directory structure
    indir = Path(tempfile.mkdtemp())

    fasta_file_1_fasta = tempfile.NamedTemporaryFile(mode='w+', dir=indir, encoding='utf-8', suffix=".fasta")
    fasta_file_1_fasta.write(">Header\nATCG")
    fasta_file_1_fasta.flush()

    fasta_file_2_fna = tempfile.NamedTemporaryFile(mode='w+', dir=indir, encoding='utf-8', suffix=".fna")
    fasta_file_2_fna.write(">Header\nATCG")
    fasta_file_2_fna.flush()

    fasta_file_3_fa = tempfile.NamedTemporaryFile(mode='w+', dir=indir, encoding='utf-8', suffix=".fa")
    fasta_file_3_fa.write(">Header\nATCG")
    fasta_file_3_fa.flush()

    sample_name_dict = get_sample_name_dict(indir=indir)
    assert len(sample_name_dict) == 3

    shutil.rmtree(indir)


def test_bulk_sample_typing():
    test_data_dir = Path(__file__).parent / 'data'
    assert test_data_dir.exists()
    test_database = Path(__file__).parent / 'data' / 'database'
    assert test_database.exists()

    # Set up temporary directory
    tmpoutdir = Path(f'/tmp/typer_testing_bulk')
    if tmpoutdir.exists():
        shutil.rmtree(tmpoutdir)

    detailed_report_list = bulk_sample_typing(indir=test_data_dir, outdir=tmpoutdir, database=test_database,
                                              keep_blast=False)
    # There should be 3 reports generated - one for each input sample
    assert len(detailed_report_list) == 3



import os
import logging
import pandas as pd
from pathlib import Path
from tqdm import tqdm


def call_sequence_comparison(targets: list, outdir: Path):
    os.makedirs(str(outdir), exist_ok=False)
    logging.debug(f"Created directory {outdir}")

    # Populate target_dict (key=sample name, value=path to report)
    targets = [Path(target) for target in targets]
    target_dict = {}
    for target in targets:
        target_name = target.with_suffix("").name.replace("_cgMLST_Allele_Report", "")
        target_dict[target_name] = target

    # Make pairwise comparisons
    pairwise_dict = {}
    mismatch_dict = {}

    logging.info("Counting mismatches between samples")
    for target_1, target_1_report in tqdm(target_dict.items()):
        inner_dict = {}
        inner_mismatch_dict = {}
        for target_2, target_2_report in target_dict.items():
            # Difference dict
            difference_dict = compare_reports(target_1_report, target_2_report)
            inner_dict[target_2] = difference_dict

            # Total mismatches
            mismatches = count_mismatches(difference_dict)
            inner_mismatch_dict[target_2] = mismatches

        mismatch_dict[target_1] = inner_mismatch_dict
        pairwise_dict[target_1] = inner_dict

    logging.info("Conducting pairwise comparisons")
    for key, val in tqdm(pairwise_dict.items()):
        df = pd.DataFrame.from_dict(val)
        out_name = outdir / (key + "_pairwise_comparisons.xlsx")
        if out_name.exists():
            continue
        writer = pd.ExcelWriter(str(out_name), engine='xlsxwriter')
        df.to_excel(writer, sheet_name=key)
        worksheet = writer.sheets[key]
        worksheet.conditional_format('B2:AZ4000', {'type': '2_color_scale'})
        writer.save()

    master_df = []
    out_name = outdir / "all_samples_comparison.xlsx"
    logging.info("Processing data for final comparison report")
    for key, val in tqdm(mismatch_dict.items()):
        df = pd.DataFrame.from_dict(val, orient='index').transpose()
        df['id'] = key
        df = df.set_index('id')
        master_df.append(df)
    df = pd.concat(master_df)

    writer = pd.ExcelWriter(str(out_name), engine='xlsxwriter')
    df.to_excel(writer, sheet_name='all_samples_comparison')
    worksheet = writer.sheets['all_samples_comparison']
    worksheet.conditional_format('B2:AZ200', {'type': '3_color_scale'})
    writer.save()


def compare_reports(report_1: Path, report_2: Path) -> dict:
    dict_1 = read_allele_report(report_1)
    dict_2 = read_allele_report(report_2)
    difference_dict = compare_allele_dicts(dict_1, dict_2)
    mismatches = count_mismatches(difference_dict)
    mismatch_pct = (mismatches / len(difference_dict)) * 100
    logging.debug(f"NUM. MISMATCHES: {mismatches}/{len(difference_dict)}")
    logging.debug(f"PERCENT MISMATCH: {mismatch_pct}%")
    return difference_dict


def read_allele_report(report: Path) -> dict:
    df = pd.read_csv(report, delimiter="\t", index_col=None, header=None)
    dct = dict(zip(df.iloc[:, 0], df.iloc[:, 1]))
    for key, val in dct.items():
        try:
            dct[key] = str(int(val))
        except ValueError:
            dct[key] = "NA"
    return dct


def compare_keys(dict_1: dict, dict_2: dict):
    keys_1 = dict_1.keys()
    keys_2 = dict_2.keys()
    error_state = False

    if len(keys_1) != len(keys_2):
        error_state = True
        logging.warning(f"WARNING: Length of dictionaries are not the same. {len(keys_1)} vs. {len(keys_2)}")

    for key in keys_1:
        if key not in keys_2:
            error_state = True
            logging.warning(f"WARNING: {key} not in both dictionaries")

    for key in keys_2:
        if key not in keys_1:
            error_state = True
            logging.warning(f"WARNING: {key} not in both dictionaries")

    if error_state:
        logging.error(f"FAIL: Provided dictionaries are not equivalent. Beware.")
    else:
        pass


def compare_allele_dicts(dict_1: dict, dict_2: dict):
    # Error check
    compare_keys(dict_1, dict_2)

    # Produce match dictionary
    difference_dict = {}
    for key, val in dict_1.items():
        if key not in dict_2.keys():
            difference_dict[key] = 1  # Missing (counts as no match)
            continue

        if dict_1[key] == dict_2[key]:
            difference_dict[key] = 0  # Match
        else:
            difference_dict[key] = 1  # No match
    return difference_dict


def count_mismatches(difference_dict: dict) -> int:
    mismatch_counter = 0
    for key, val in difference_dict.items():
        mismatch_counter += val
    return mismatch_counter



import setuptools
from EnteroTyper.__init__ import __version__, __author__, __email__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="EnteroTyper",
    install_requires=['click', 'pandas', 'xlsxwriter', 'tqdm', 'dataclasses'],
    python_requires='~=3.6',
    description="Scripts to type and compare assemblies based off of Enterobase schemes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BFSSI-Bioinformatics-Lab/EnteroTyper",
    package_data={'EnteroTyper': ['*']},
    install_package_data=True,
    packages=setuptools.find_packages(),
    version=__version__,
    author=__author__,
    author_email=__email__,
    entry_points={
        'console_scripts': [
            'enterotyper=EnteroTyper.enterotyper:enterotyper'
        ]
    }
)



import click
import logging
from pathlib import Path
from EnteroTyper.bin.typer import type_sample, makeblastdb_database
from EnteroTyper.bin.bulk_typer import bulk_sample_typing
from EnteroTyper.bin.concatenate_sequences import sequence_concatenation_pipeline
from EnteroTyper.bin.compare_sequences_types import call_sequence_comparison
from EnteroTyper.bin.accessories import check_all_dependencies

from EnteroTyper.__init__ import __email__, __author__, __version__

DEPENDENCIES = [
    'blastn',
    'blastx',
    'makeblastdb',
    'muscle'
]


def convert_to_path(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    value = Path(value)
    return value


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    logging.info(f"Version: {__version__}")
    logging.info(f"Author: {__author__}")
    logging.info(f"Email: {__email__}")
    quit()


@click.group()
def enterotyper():
    pass


@enterotyper.command(help="Types an assembly based on an Enterobase scheme",
                     short_help="Types an assembly based on an Enterobase scheme")
@click.option('-i', '--input-assembly',
              type=click.Path(exists=True),
              required=True,
              default=None,
              help='Path to input assembly in FASTA format',
              callback=convert_to_path)
@click.option('-db', '--database',
              type=click.Path(exists=True),
              required=True,
              default=None,
              help='Path to your MLST database',
              callback=convert_to_path)
@click.option('-o', '--outdir',
              type=click.Path(exists=False),
              required=True,
              default=None,
              help='Root directory to store all output files',
              callback=convert_to_path)
@click.option('--create-db',
              help='Set this flag to create the blastDB files using makeblastdb in the specified database directory.'
                   'Will re-create the database files if they are already present.',
              is_flag=True,
              required=False,
              default=False)
@click.option('--keep-blast',
              help='Set this flag to keep all BLASTn output generated by EnteroTyper.',
              is_flag=True,
              required=False,
              default=False)
@click.option('-v', '--verbose',
              is_flag=True,
              default=False,  # Set this to false eventually
              help='Set this flag to enable more verbose logging.')
def typer(input_assembly: Path, database: Path, outdir: Path, create_db: bool, keep_blast: bool, verbose: bool):
    if verbose:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.DEBUG,
            datefmt='%Y-%m-%d %H:%M:%S')
    else:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.INFO,
            datefmt='%Y-%m-%d %H:%M:%S')

    logging.info("Starting Enterotyper typer script")
    if not check_all_dependencies(DEPENDENCIES):
        quit()
    type_sample(input_assembly=input_assembly, database=database,
                outdir=outdir, keep_blast=keep_blast, create_db=create_db)
    logging.info("Script complete")


@enterotyper.command(help="Runs typer on many samples simultaneously",
                     short_help="Runs typer on many samples simultaneously")
@click.option('-i', '--indir',
              type=click.Path(exists=True),
              required=True,
              default=None,
              help='Path to directory containing FASTA assemblies',
              callback=convert_to_path)
@click.option('-db', '--database',
              type=click.Path(exists=True),
              required=True,
              default=None,
              help='Path to your MLST database',
              callback=convert_to_path)
@click.option('-o', '--outdir',
              type=click.Path(exists=False),
              required=True,
              default=None,
              help='Root directory to store all output files',
              callback=convert_to_path)
@click.option('--create-db',
              help='Set this flag to create the blastDB files using makeblastdb in the specified database directory.'
                   'Will re-create the database files if they are already present.',
              is_flag=True,
              required=False,
              default=False)
@click.option('--keep-blast',
              help='Set this flag to keep all BLASTn output generated by EnteroTyper.',
              is_flag=True,
              required=False,
              default=False)
@click.option('-v', '--verbose',
              is_flag=True,
              default=False,  # Set this to false eventually
              help='Set this flag to enable more verbose logging.')
def bulk(indir: Path, database: Path, outdir: Path, create_db: bool, keep_blast: bool, verbose: bool):
    if verbose:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.DEBUG,
            datefmt='%Y-%m-%d %H:%M:%S')
    else:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.INFO,
            datefmt='%Y-%m-%d %H:%M:%S')
    logging.info("Starting Enterotyper bulk script")
    if not check_all_dependencies(DEPENDENCIES):
        quit()

    # Set up the database if needed
    if create_db:
        makeblastdb_database(database=database)

    # Type every sample
    detailed_report_list = bulk_sample_typing(indir=indir, outdir=outdir, database=database, keep_blast=keep_blast)

    # Concatenate sequences
    sequence_concatenation_pipeline(targets=detailed_report_list,
                                    outdir=outdir / 'concatenated_sequences',
                                    database=database)
    logging.info("Script complete")


@enterotyper.command(short_help="Concatenate sequences from report files generated by typer",
                     help="Takes a list of target *.BLASTn_Detailed_Report.tsv files as input. "
                          "Extracts sequences from the BLASTn report files as FASTA files, aligns them all with "
                          "MUSCLE, and then concatenates all sequences into a single FASTA. ")
@click.option('-o', '--outdir',
              type=click.Path(exists=False),
              required=True,
              default=None,
              help='Root directory to store all output files',
              callback=convert_to_path)
@click.option('-db', '--database',
              type=click.Path(exists=True),
              required=True,
              default=None,
              help='Path to your MLST database',
              callback=convert_to_path)
@click.option('-v', '--verbose',
              is_flag=True,
              default=False,  # Set this to false eventually
              help='Set this flag to enable more verbose logging.')
@click.argument('targets', nargs=-1, type=click.Path(exists=True))
def concatenate(targets, outdir, database, verbose):
    if verbose:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.DEBUG,
            datefmt='%Y-%m-%d %H:%M:%S')
    else:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.INFO,
            datefmt='%Y-%m-%d %H:%M:%S')

    logging.info("Starting Enterotyper concatenate script")
    if not check_all_dependencies(DEPENDENCIES):
        quit()
    sequence_concatenation_pipeline(targets=targets, database=database, outdir=outdir)
    logging.info("Script complete")


@enterotyper.command(short_help="Compares sequence types across samples processed by typer",
                     help="Takes a list of target *.cgMLST_Allele_Report.tsv files as input. "
                          "Number of mismatches will be indicated in output files - a zero means no mismatches "
                          "between types.")
@click.option('-o', '--outdir',
              type=click.Path(exists=False),
              required=True,
              default=None,
              help='Root directory to store all output files',
              callback=convert_to_path)
@click.option('-v', '--verbose',
              is_flag=True,
              default=False,
              help='Set this flag to enable more verbose logging.')
@click.argument('targets', nargs=-1, type=click.Path(exists=True))
def compare(targets, outdir, verbose):
    if verbose:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.DEBUG,
            datefmt='%Y-%m-%d %H:%M:%S')
    else:
        logging.basicConfig(
            format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
            level=logging.INFO,
            datefmt='%Y-%m-%d %H:%M:%S')
    logging.info("Starting Enterotyper compare script")
    if not check_all_dependencies(DEPENDENCIES):
        quit()
    call_sequence_comparison(targets=targets, outdir=outdir)
    logging.info("Script complete")


if __name__ == "__main__":
    cli = click.CommandCollection(sources=[enterotyper],
                                  help="Enterotyper is a suite of tools to type and evaluate assemblies with "
                                       "Enterobase typing schemes.")
    cli()



import tempfile
from EnteroTyper.bin.typer import *
from pathlib import Path


def test_type_sample():
    """
    Uses dummy data in the test_data_dir directory. Uses the MLST_Achtman database.
    assembly_1.fasta should contain only recA_1 and fumC_7
    assembly_2.fasta should contain only recA_1 and fumC_16
    assembly_3.fasta should not contain any markers
    :return:
    """
    # Establish test data directories, make sure they are structured correctly
    test_data_dir = Path(__file__).parent / 'data'
    assert test_data_dir.exists()
    test_database = Path(__file__).parent / 'data' / 'database'
    assert test_database.exists()

    test_assembly_1 = test_data_dir / 'assembly_1.fasta'  # recA_1 and fumC_7
    test_assembly_2 = test_data_dir / 'assembly_2.fasta'  # recA_1 and fumC_16
    test_assembly_3 = test_data_dir / 'assembly_3.fasta'  # No markers
    test_assemblies = [test_assembly_1, test_assembly_2, test_assembly_3]

    # Test if reports are generated
    for assembly in test_assemblies:
        assert assembly.exists()
        # Set up temporary directory
        tmpoutdir = Path(f'/tmp/typer_testing_{assembly.name}')
        if tmpoutdir.exists():
            shutil.rmtree(tmpoutdir)
        # Run typer
        detailed_report = type_sample(input_assembly=assembly, database=test_database,
                                      outdir=tmpoutdir, create_db=True, keep_blast=False)
        # Ensure report was generated
        assert detailed_report.exists()

        # Verify contents of report
        df = pd.read_csv(detailed_report, sep="\t")
        expected_columns = ['qseqid', 'sseqid', 'slen', 'length', 'qstart', 'qend',
                            'pident', 'lratio', 'locus', 'hit_type', 'qseq_strand_aware']
        if 'assembly_3' not in assembly.name:
            assert list(df.columns) == expected_columns

        # Verify individual assembly test case results
        if 'assembly_1' in assembly.name:
            assert 'fumC_7' in df['sseqid'].tolist()
            assert 'recA_1' in df['sseqid'].tolist()
        elif 'assembly_2' in assembly.name:
            assert 'fumC_16' in df['sseqid'].tolist()
            assert 'recA_1' in df['sseqid'].tolist()
        elif 'assembly_3' in assembly.name:
            hit_types = list(df['hit_type'].unique())
            assert len(hit_types) == 1
            assert hit_types[0] == 'NO_MATCH'

        # Remove temporary directory
        shutil.rmtree(tmpoutdir)


def test_move_blastn_files():
    # Create mock directory structure
    outdir = Path(tempfile.mkdtemp())
    indir = Path(tempfile.mkdtemp())
    blastn_file = tempfile.NamedTemporaryFile(mode='w+', dir=indir, encoding='utf-8', suffix=".BLASTn")
    blastn_file.write("tmp")
    blastn_file.flush()

    # Ensure indir has BLASTn file inside of it
    assert len((list(indir.glob("*.BLASTn")))) != 0

    # Call method on indir and confirm that BLASTn file was moved to outdir
    move_blastn_files(indir=indir, outdir=outdir)
    assert len((list(outdir.glob("*.BLASTn")))) != 0

    # Cleanup
    shutil.rmtree(outdir)
    shutil.rmtree(indir)


def test_delete_blastn_files():
    # Create mock directory structure
    indir = Path(tempfile.mkdtemp())
    blastn_file = tempfile.NamedTemporaryFile(mode='w+', dir=indir, encoding='utf-8', suffix=".BLASTn")
    blastn_file.write("tmp")
    blastn_file.flush()

    # Ensure indir has BLASTn file inside of it
    assert len((list(indir.glob("*.BLASTn")))) != 0
    delete_blastn_files(indir=indir)

    # Ensure file was deleted
    assert len((list(indir.glob("*.BLASTn")))) == 0


def test_generate_cgmlst_report():
    pass


def test_generate_detailed_report():
    pass


def test_get_database_files():
    pass


def test_multiprocess_blastn_call():
    pass


def test_closest_allele_df():
    pass


def test_get_sequence_type():
    pass


def test_process_blastn_df():
    pass


def test_call_makeblastdb():
    pass


def test_makeblastdb_database():
    pass


def test_get_allele_length():
    pass


def test_call_blastn():
    pass


def test_combine_dataframes():
    pass


def test_parse_blastn():
    pass


def test_get_reverse_complement_row():
    pass



from pathlib import Path
from EnteroTyper.bin.typer import type_sample


def get_sample_name_dict(indir: Path) -> dict:
    fasta_files = list(indir.glob("*.fna"))
    fasta_files += list(indir.glob("*.fasta"))
    fasta_files += list(indir.glob("*.fa"))

    sample_name_dict = {}
    for f in fasta_files:
        sample_name = f.with_suffix("").name
        sample_name_dict[sample_name] = f
    return sample_name_dict


def bulk_sample_typing(indir: Path, outdir: Path, database: Path, keep_blast: bool) -> list:
    sample_name_dict = get_sample_name_dict(indir=indir)
    detailed_report_list = []
    for sample_name, assembly in sample_name_dict.items():
        sample_out_dir = outdir / sample_name
        detailed_report = type_sample(input_assembly=assembly, database=database, outdir=sample_out_dir,
                                      create_db=False, sample_name=sample_name, keep_blast=keep_blast)
        detailed_report_list.append(detailed_report)
    return detailed_report_list
